import asyncio
import threading

from loguru import logger

from ksrpc.connections import BaseConnection


class RpcClient:
    """
    RpcClient根据调用生成网络请求

    Warnings
    --------
    先调用一到多个`__getattr__`，然后`__call__`收集`__getattr__`的结果。
    并行时`__getattr__`的调用先后不可控，所以不能在`asyncio.gather`中使用

    """

    def __init__(self,
                 module: str,
                 connection: BaseConnection,
                 ):
        """初始化

        Parameters
        ----------
        module: str
            顶层模块名
        connection: Connection
            连接对象

        """
        self._module = module
        self._connection = connection
        self._names = []
        self._lock = threading.Lock()

    def __getattr__(self, name):
        with self._lock:
            # 同一对象在asyncio.gather中使用会混乱，请用独立对象
            self._names.append(name)
            return self

    @property
    def __doc__(self):
        # 比较特殊，默认是调用`RpcClient.__doc__`，这样写才行
        return self.__getattr__('__doc__')

    def __del__(self):
        self._connection = None

    async def __call__(self, *args, **kwargs):
        with self._lock:
            name = '.'.join(self._names)
            # 用完后得重置，否则第二次用时不正确了
            self._names = []

        # 排序，参数顺序统一后，排序生成key便不会浪费了
        kwargs = dict(sorted(kwargs.items()))
        try:
            return await self._connection.call(self._module, name, args, kwargs)
        except asyncio.TimeoutError:
            await self._connection.reset()  # 重置
            logger.warning(f'{self._module}::{name} timeout')
            raise
        except Exception as e:
            await self._connection.reset()  # 重置
            logger.warning(f'{self._module}::{name} error, {e}')
            raise


class RpcProxy:
    """每次调用都生成一个`RpcClient`对象，占用资源略多于`RpcClient`

    Notes
    --------
    `__getattr__`后就已经是一个独立的`RpcClient`对象，可以在`asyncio.gather`中并行使用

    """

    def __init__(self,
                 module: str,
                 connection: BaseConnection,
                 ):
        """初始化

        Parameters
        ----------
        module: str
            顶层模块名
        connection: Connection
            连接对象
        """
        self._module = module
        self._connection = connection

    def __getattr__(self, name):
        # 第一个方法调用用来生成新RpcClient对象，用来解决不能并发的问题
        return RpcClient(self._module, self._connection).__getattr__(name)

    @property
    def __doc__(self):
        return self.__getattr__('__doc__')

    async def __call__(self, *args, **kwargs):
        return await RpcClient(self._module, self._connection)()

    def __del__(self):
        self._connection = None



class BaseConnection:
    def __init__(self, url):
        self._url = url

    def get_url(self):
        if isinstance(self._url, str):
            return self._url
        else:
            # 可以定制URL，比如随机生成与本地时间有关的url
            return self._url()

    async def reset(self):
        ...

    async def call(self, module, name, args, kwargs):
        ...



import asyncio
import hashlib
import inspect
import sys
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime
from importlib import import_module

from loguru import logger

from ksrpc.config import CALL_IN_NEW_PROCESS

logger.remove()
logger.add(sys.stderr,
           format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level:<8}</level> | PID:{process.id} | TID:{thread.id} | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>",
           level="INFO", colorize=True)


def make_key(module, name, args, kwargs):
    """生成缓存key"""
    key = f'{module}::{name}_{repr(args)}_{repr(kwargs)}'
    return hashlib.md5(key.encode('utf-8')).hexdigest()


def get_func(module, name):
    # ksrpc.server.demo.CLASS.B.__getitem__(3)，这种格式只能确信第一个是模块，最后一个是方法，其他都不确定
    m = import_module(module)
    for n in name.split('.'):
        if len(n) == 0:
            continue
        if hasattr(m, n):
            m = getattr(m, n)
        else:
            m = import_module(f"{m.__name__}.{n}")
    return m


async def async_call(module, name, args, kwargs):
    """简版异步API调用。没有各种额外功能"""
    key = make_key(module, name, args, kwargs)

    # 返回的数据包
    d = dict(status=200,  # status.HTTP_200_OK,
             datetime=datetime.now().isoformat(),  # 加查询时间，缓存中也许可以判断是否过期
             module=module,
             name=name,
             args=args,
             kwargs=kwargs)
    try:
        # 转成字符串，后面可能于做cache的key
        logger.info(f'{module}::{name}\t{args}\t{kwargs}'[:200])

        func = get_func(module, name)

        # 可以调用的属性
        if callable(func):
            # 例如os.remove
            if inspect.iscoroutinefunction(func):
                output = await func(*args, **kwargs)
            else:
                output = func(*args, **kwargs)
        else:
            output = func

        # 结果的类型名
        class_name = output.__class__.__name__

        # 不可调用的直接返回
        if class_name == 'module':
            # 例如os.path，但用法为os.path()
            data = repr(output)
        elif class_name in ('dict_keys', 'dict_values'):
            # 无法序列化，只能强行转换
            # 例如：sys.modules.keys()
            data = list(output)
        else:
            # 例如math.pi，但得加()才能触发，math.pi()
            data = output

        d['type'] = type(output).__name__
        d['data'] = data
    except Exception as e:
        d['status'] = 500  # status.HTTP_500_INTERNAL_SERVER_ERROR
        d['type'] = type(e).__name__
        d['data'] = repr(e)

    return key, d


def async_wrapper(func, *args, **kwargs):
    try:
        # 尝试使用 Python 3.7+ 的 asyncio.run()
        return asyncio.run(func(*args, **kwargs))
    except AttributeError:
        # 回退到手动事件循环管理
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            return loop.run_until_complete(func(*args, **kwargs))
        finally:
            loop.close()


async def process_call(module, name, args, kwargs):
    """进程版API调用。结束后进程退出自动释放内存

    Notes
    =====
    启动了一个新进程速度显著下降，但与网络传输大数据相比可以忽略不计。
    获得好处是释放了内存，适合内存有限的平台

    """
    with ProcessPoolExecutor(max_workers=1) as executor:
        future = executor.submit(async_wrapper, async_call, module, name, args, kwargs)
        return future.result()


async def switch_call(module, name, args, kwargs):
    if CALL_IN_NEW_PROCESS:
        return await process_call(module, name, args, kwargs)
    else:
        return await async_call(module, name, args, kwargs)



import base64
import pickle
import zlib

from aiohttp import web

from ksrpc.caller import switch_call
from ksrpc.config import USER_CREDENTIALS, URL_CHECKER
from ksrpc.utils.chunks import send_in_chunks, data_sender


@web.middleware
async def basic_auth_middleware(request, handler):
    # 检查是否需要跳过认证的路由
    if request.path == '/public':
        return await handler(request)

    # 获取 Authorization 头
    auth_header = request.headers.get('Authorization')

    if auth_header is None or not auth_header.startswith('Basic '):
        return unauthorized_response()

    # 解码凭证
    try:
        auth_bytes = base64.b64decode(auth_header[6:])
        auth_str = auth_bytes.decode('utf-8')
        username, password = auth_str.split(':', 1)
    except (ValueError, UnicodeDecodeError):
        return unauthorized_response()

    # 验证凭证
    if USER_CREDENTIALS.get(username) != password:
        return unauthorized_response()

    # 认证通过，添加用户名到请求对象
    request['user'] = username
    return await handler(request)


def unauthorized_response():
    return web.Response(
        status=401,
        headers={'WWW-Authenticate': 'Basic realm="Restricted Area"'},
        text="Unauthorized"
    )


@web.middleware
async def url_check_middleware(request, handler):
    try:
        URL_CHECKER.check(request)
    except (AssertionError, ValueError):
        return web.HTTPForbidden()

    return await handler(request)


async def handle(request: web.Request) -> web.StreamResponse:
    buffer = bytearray()
    buf = bytearray()
    async for chunk, end_of_http_chunk in request.content.iter_chunks():
        buf.extend(chunk)
        if end_of_http_chunk:
            if len(buf) == 0:
                continue
            buffer.extend(zlib.decompress(buf))
            buf.clear()

    key, data = await switch_call(**pickle.loads(buffer))
    buffer.clear()

    body = pickle.dumps(data)
    headers = {'Content-Disposition': f"{key}.pkl.chunked.zip"}

    del data
    return web.Response(body=data_sender(body, print), headers=headers)


async def websocket_handler(request: web.Request) -> web.StreamResponse:
    ws = web.WebSocketResponse()
    await ws.prepare(request)

    buffer = bytearray()
    async for msg in ws:
        if msg.type is web.WSMsgType.BINARY:
            buffer.extend(zlib.decompress(msg.data))
        elif msg.type == web.WSMsgType.TEXT:
            if msg.data == "EOF":
                key, data = await switch_call(**pickle.loads(buffer))
                buffer.clear()
                await send_in_chunks(ws, pickle.dumps(data), print)
                del data
        elif msg.type == web.WSMsgType.ERROR:
            print('Server WebSocket connection closed with exception %s' % ws.exception())
        elif msg.type is web.WSMsgType.CLOSE:
            print('Server WebSocket connection closed')
            break
    print("End of websocket_handler")
    return ws


def sync_app(argv):
    # uv run python -m aiohttp.web -H 0.0.0.0 -P 8080 ksrpc.run_app:sync_app
    app = web.Application(middlewares=[
        basic_auth_middleware,  # 注释此行屏蔽Baisc认证
        url_check_middleware,
    ])
    app.add_routes([
        web.post("/api/{path}", handle),
        web.get("/ws/{path}", websocket_handler),
    ])
    return app


async def async_app(argv):
    # gunicorn ksrpc.run_app:async_app --bind 0.0.0.0:8080 --worker-class aiohttp.GunicornWebWorker
    app = sync_app(argv)
    return app



import time

from aiohttp import web


class BaseURL:
    """
    为防路径扫描，使用动态路径
    """

    def __init__(self, url: str = 'http://127.0.0.1:8080/api/file'):
        self._url = url
        self._path = self._url.split('?', 1)[0]
        self._path = self._path.rsplit('/', 1)[1]
        pass

    def __call__(self, *args, **kwargs):
        """由客户端使用"""
        return self._url

    def check(self, request: web.Request):
        """由服务端使用"""
        path = request.match_info.get('path', '')
        assert path == self._path


class TimeURL(BaseURL):
    """时间动态路径"""

    def __init__(self, url: str = 'ws://127.0.0.1:8080/ws/{}'):
        super().__init__(url)

    def __call__(self, *args, **kwargs):
        """由客户端使用"""
        url = self._url.format(int(time.time()))
        print('Use URL:', url)
        return url

    def check(self, request: web.Request):
        """由服务端使用"""
        path = request.match_info.get('path', '0')
        assert abs(time.time() - int(path)) < 30
